# 八、`super`和`this`关键字和异常在 Java 中的重要性

在本章中，我们将了解两个关键词：`super`和`this`。在编写 Java 代码时，我们将挑选一些示例并解释如何在各种情况下使用它们。我们还将研究异常，以及如何使用它们来处理由于某些错误导致代码失败的情况。我们将在`finally`部分结束本章。

在本章中，我们将介绍以下主题：

*   `super`关键字
*   `super`关键字的实际使用
*   这个关键词的重要性
*   不同种类的例外
*   处理异常的`try…catch`机制
*   `finally`块在 Java 中的重要性

# `super`关键字

通常，当人们从不同的类继承属性时，如果在父类和子类中使用相同的变量名，则可能存在冗余。为了区分父变量和子变量，我们使用了`super`关键字。

让我们用一个例子来解释这一点。让我们创建两个类，并将它们命名为`childDemo`和`parentDemo`。在`parentDemo`类中，我们定义了一个名为`name`的字符串，并将`'rahul'`字符串分配给它。

现在，在`childDemo`类中，我们继承了`parentDemo`的属性。我们知道如何使用`extends`关键字继承父类的属性，这是我们在第 5 章中学习的关于接口和继承的所有知识。此处给出了继承属性的代码：

```java
public class childDemo extend parentDemo{
```

在此代码中，`childDemo`取`parentDemo`的属性。

在`childDemo`类中添加一个字符串，将其命名为`name`，并将`QAClickAcademy`字符串分配给它。然后，我们在`childDemo`类中定义一个名为`public void getStringdata()`的方法，并给出一条语句来打印`name`的值作为输出。我们在`getStringdata()`之外定义了另一个名为`public static void main(String[] args)`的方法，并为子类`childDemo cd = new childDemo();`创建了一个对象。一旦对象被创建，我们会在下面添加另一行代码：`cd.getStringdata();`。这就调用了`getrStringdata()`方法，因此很明显，名称被打印为输出，即`QAClickAcademy`。尽管我们继承了`parentDemo`类的属性，该类也包含同名字符串，但`print`语句调用`childDemo`中字符串的值。这是因为 Java 优先考虑局部变量。

每当父级和子级的变量名发生冲突时，它会优先选择局部变量，在本例中是`childDemo`类。如果我们正在进行一个项目，需要在`parentDemo`类中打印字符串名称，该怎么办？为此，我们使用`super`关键字引用`parentDemo`类，从中我们将属性继承到`childDemo`类。因此，如果我们想从`parentDemo`类中调用`name`变量，我们将添加一个`print`语句，并在我们想要打印的变量之前添加一个`super`关键字，这将从`parentDemo`中获得值。当我们现在运行代码时，我们得到父对象和子对象作为输出，因为我们在两个类中都为名称字符串留下了一个`print`语句。`parentDemo`类的代码如下：

```java
public class parentDemo{
    String name= "Rahul";
    public static viod main (String[] args){
    }
}
```

`childDemo`类的代码如下：

```java
public class childDemo extends parentDemo{

    String name = "QAClickAcademy";
    public void getStringdata();
    {
        System.out.println(name);
        System.out.println(super.name);
    }

    public static void main(String[] args){
    childDemo cd = new childDemo();
    cd.getStringdata();
    }
}
```

最终结果将是：

```java
QAClickAcademy
Rahul
```

# `super`关键字的实际使用

在本节中，我们将了解在 Java 中使用`super`关键字的不同方式。

# 对方法使用`super`关键字

我们看到了如何通过`super`关键字处理父变量。在本节中，我们将了解如何处理两个方法，如果它们在`parentDemo`和`childDemo`类中的名称相同。在本节中，我们也将使用前面的示例。

在`parentDemo`类中，添加一个名为`getData()`的方法，并在该方法内部给出一条打印语句来显示`"I am in parent class"`消息。如果我们想在`childDemo`类中执行`getData()`方法，我们就在`childDemo`类的`main`方法中写入`cd.getData()`。我们可以访问`getData()`，因为我们继承了`parentDemo`类的属性。如果我们运行`childDemo`类，我们将收到前面示例的输出以及我们在`parentDemo`类`I am in parent class`中添加的新句子。

在`childDemo`类中，我们将定义另一个与`parentDemo`类同名的方法，并添加一个`print`语句来显示`I am in child class`消息。如果我们运行`childDemo`类，我们将获得前面示例的输出，然后显示`I am in child class`。这是因为优先考虑了本地类，所以`childDemo`类中的`getData()`方法会覆盖`parentDemo`类中的`getData()`方法。

现在，我们想在`childDemo`类中使用`parentDemo`类的`getData()`方法。要做到这一点，我们只需像对待变量那样：在`childDemo`类的`getData()`方法中添加`super.getData()`。当我们运行`childDemo()`类时，我们得到前面示例的输出，然后是`I am in parent class`和`I am in child class`。

# 为构造函数使用`super`关键字

让我们在本节中为构造函数使用`super`关键字。这里我们也将使用前面的示例。

在`parentDemo`类中，我们定义了一个构造函数`parentDemo()`，并添加了一个打印语句来打印：`Parent class constructor`。

在`childDemo`中，我们定义了一个构造函数`childDemo()`，并添加了一个打印语句来打印：`Child class constructor`。如果我们想在`childDemo`类中使用`parentDemo`类构造函数，我们会在`childDemo()`构造函数中添加`super()`方法。这使得控制器调用`parentDemo`类中的构造函数。

在使用构造函数时，我们需要遵循一条重要的规则：无论何时在子构造函数中使用`super`构造函数，它都应该始终位于第一行。

当我们运行`childDemo`类时，控制器首先执行`super()`方法。它进入`parentDemo()`构造函数并执行它，然后进入`childDemo()`。因此，最终输出将是：

```java
Parent class constructor
Child class constructor
QAClickAcademy
Rahul
I am parent class
I am in child class
```

# 这个关键词的重要性

Java 中还有一个类似于`super`关键字的关键字：`this`。在本节中，我们将看一看`this`关键字。

让我们用一个例子来解释`this`关键字。创建一个名为`thisDemo`的类，声明一个变量`a`，并为其赋值`2`。我们在其类中定义了一个`getData()`方法，在其中声明了`a`变量，并为其赋值`3`。我们还添加了一个`print`语句。代码如下所示：

```java
package coreJava;public class thisDemo
{
    int a= 2;
    public void getData()
    {
        int a= 3;
        System.out.println(a);
    }
```

正如我们所看到的，`a`的值在整个类中是`2`，但是在一个具体的方法`getData()`中，我们希望变量的值是`3`。在这段代码中，我们希望调用`a`的两个值，即`2`和`3`。我们在`main`方法中创建了一个对象，并在其中添加了`td`对象。`td`对象的代码如下：

```java
thisDemo td=new thisDemo();
td.getData();
```

如果我们运行代码，我们得到的输出是`3`。但我们也希望在同一块中将`a`的值打印为`2`。此时`this`关键字开始发挥作用。类对象的范围将在类级别，而不是方法级别。所以我们说`getData()`方法指的是当前对象，对象范围在类级别。所以`a= 2`对整个类有效，`a=3`只对`getData()`方法有效。这就是为什么我们将`getData()`方法中的`a`变量称为局部变量，将类中的`a`变量称为全局变量。

要打印我们正在处理的示例的全局变量，我们需要在`getData()`方法中添加一个`print`语句，并在`print`语句中添加`this.a`。`print`语句如下所示：

```java
System.out.println(this.a);
```

当我们运行代码时，我们得到以下输出：

```java
3
2
```

这就结束了我们关于`this`变量的示例。现在让我们了解异常。

# 不同种类的例外

在本节中，我们将了解如何在 Java 中处理异常。

一般来说，如果代码中有错误，我们需要捕获它并打印一条消息而不会失败；这可以通过`try...catch`机制实现。所以一般来说，当我们试图编写代码时，我们怀疑其中可能有错误，我们将使用该错误进行异常处理。

我们将通过一个练习来解释它。让我们创建一个新类`exceptionDemo`，在`main`块中声明`a`、`b`和`c`变量，并分别为它们分配`4`、`7`和`0`的值。我们在主块中添加一个`try`块，并声明一个整数变量`k`，它等于`b`除以`c`。每当我们在`try`块中添加任何内容时，我们都会尝试查看代码是否有效。如果失败，控制器将从该`try`块中出来并进入包含异常的`catch`块。需要记住的一点是，`catch`块正好位于`try`块之后。在`catch`块中，我们编写一条打印消息来显示`I caught the error/exception`。

当控制器进入`k`变量行时，脚本失败，因为`7/0`为无穷大，这是一个算术异常，但脚本不会立即失败。如果我们不写`try...catch`块，我们会看到另一种错误。

让我们取出`try...catch`块，运行代码，看看我们得到的错误。我们在输出部分`Java.lang.ArithmeticException`中看到一个错误；这是因为我们无法将`7`除以`0`，所以脚本突然失败。

如果我们最初觉得我们的代码中会出现错误，我们可以简单地创建一个脚本来传递并捕获错误，方法是放置一条适当的调试消息，该消息可以在`try...catch`机制的帮助下处理。现在，让我们再次添加`try...catch`块并调试整个代码。输出为`I caught the error/exception`；这是因为`7`除以`0`是无穷大，所以这里的脚本应该失败，但我们在输出部分没有看到任何错误，表示代码失败。这是因为控制器只是移动到`catch`块并执行它。最终代码如下所示：

```java
public static void main(String[] args)
{
    int b=7; 
    int c=0;
    try
    {
        int k=b/c;
        System.out.println(k);
    }
    catch(Exception e)
    {
        System.out.println("I caught the error/exception")
    }
}
```

输出结果如下：

```java
I caught the error/exception
```

# 处理异常的`try…catch`机制

在本节中，我们将使用一个`try`后跟多个`catch`块。Java 中有不同类型的异常，我们可以为每个异常添加单独的`catch`块。

让我们用前面的例子来解释这一点。为前面代码编写的异常是一般异常，因此对于`try`块中的任何错误，都会执行一般异常。现在，让我们尝试捕获一个特定的异常。我们可以在`try`块下添加一个`catch`块，并添加一个特定的异常和一个打印语句来打印`I caught the Arithmeticerror/exception`。特定捕捉块的代码为：

```java
catch(arithmeticException et)
{
    System.out.println("I caught the Arithmeticerror/exception");
}
```

当我们运行代码时，我们得到以下输出：

```java
I caught the Arithmeticerror/exception
```

我们看到，当我们运行代码时，控制器转到`catch`块，因为`catch`块是专门为算术异常编写的，抛出的错误也属于算术错误。因此，一旦控制器收到错误，`try`块将看到哪种`catch`块与其相关，并运行它。

Java 中还有许多其他的例外：我们可以通过 Google 来查看它们。

# `finally`块在 Java 中的重要性

还有一个类似于`try...catch`块的块：`finally`块。无论是否引发异常，`finally`块都将被执行。如果程序成功运行，则执行此块；如果程序未运行，则执行此块。

我们将使用我们在“使用`try…catch`机制来处理异常”中使用的示例来解释这一点。我们只是在`catch`块之后添加一个`finally`块，并在其中给出一个打印语句，上面写着`delete cookies`。代码块如下所示：

```java
finally
{
    System.out.println("delete cookie")
}
```

当我们运行代码时，我们得到以下输出：

```java
I caught the Arithmeticerror/exception
delete cookie
```

重要的一点是，`finally`可以使用或不使用`catch`块；它只需要写在一个`try`块下面。

# 总结

在本章中，我们看了一下`super`和`this`关键字。我们还看了一些例子来解释我们可以在哪里使用这些关键词来克服某些障碍。我们了解了异常，并在代码因错误而失败的各种实例中实现了异常。我们还了解了`finally`区块。

在下一章中，我们将深入探讨集合框架，它由接口和类组成。我们还将关注三大系列：`List`、`Set`和`Map`