# 七、更加面向对象的 Java

在本章中，我们将通过创建一个超类和子类，并使用覆盖、数据结构、抽象方法和受保护方法等概念来理解它们之间的“is-a”关系，从而探索 Java 中的继承主题。我们还将深入研究抽象类的概念。

我们将详细介绍以下概念：

*   遗产
*   摘要

# 遗产

与其从一个高层次的描述开始，我认为最好是我们直接进入一个问题。

为了让我们开始，我创建了一个基本的 Java 程序，我们可以从给定的代码文件访问它。在这个程序中，我们声明了两个 Java 类：一个`Book`类和一个`Poem`类。`Book`和`Poem`类都存储了许多属性；例如，书可以有标题、作者、出版商和流派。它将所有这些属性作为构造函数输入，并提供一个`public`方法；我们可以在主程序中使用`Print`方法打印出我们所创建的任何书籍的信息。

诗的方法做了一些非常相似的事情。它有两个属性和一个`Print`方法，我们通过它的构造函数设置它的属性。我已经创建了一个非常快速的主函数来利用`Book`和`Poem`类。此函数创建一本新书和一首诗，然后将其打印出来：

```java
package inheritance;
public class Inheritance {
    public static void main(String[] args) {
        Book a = new Book(
                "The Lord Of The Rings", 
                "J.R.R. Tolkein",
                "George Allen and Unwin", 
                "Fantasy");
        Poem b = new Poem(
                "The Iliad",
                "Homer",
                "Dactylic Hexameter");

        a.Print();
        b.Print();
    }
}
```

前面的程序工作得很好，但它比需要的复杂得多。

如果我们并排看一下我们的`Book`和`Poem`类，只看它们的成员变量，我们会发现`Book`和`Poem`都有两个成员变量，即`title`和`author`：

![](img/b9086a15-f8b6-4599-808e-fdf716624758.png)

它们对成员变量执行的操作，即将它们打印到屏幕上，在这两个类中以非常相似的方式执行和实现：

![](img/4816972c-9adc-4bb7-91d4-3ca04ae260cd.png)

`Book`和`Poem`继承自一个公共类，这是一个好迹象。当我们将书籍和诗歌视为它们所代表的实物时，我们就很容易看到这一点。我们可以说，书籍和诗歌都是文学形式。

# 创建超类

一旦我们得出结论，书籍和诗歌具有某些基本属性，即所有文学的属性，我们就可以开始将这些类别分解为组成部分。例如，我们的`Book`类有两个实变量。它有一个`title`变量和一个`author`变量，这是我们与所有文献相关联的属性。它也有一个 T3 的变量和一个 T4 的变量，它可能不是唯一的书，我们不一定认为所有形式的文学都有。那么我们如何利用这些知识呢？好吧，我们可以建立我们的`Book`和`Poem`课程，让他们在基本层面上分享文学作品的本质。但要做到这一点，我们首先要教我们的节目什么是文学作品。以下是执行此操作的分步程序：

1.  我们将创建一个全新的类，并将其命名为`Literature`。
2.  我们将为这个类指定文献片段共享的属性，到目前为止我们已经声明了这些属性。在我们的例子中，书籍和诗歌已经被宣布为作品，有共同的标题和作者。所有的文学作品都有一个标题和一个作者，这样的说法在逻辑上是合理的：

```java
package inheritance;
public class Literature {
    protected String title;
    protected String author;
```

3.  从这里开始，我们将像其他人一样充实我们的`Literature`课程。我们将给它一个构造函数；在这种情况下，我们的构造函数将接受两个变量：`title`和`author`。然后，我们将它们分配给字段，就像我们在`Poem`和`Book`类中所做的那样：

```java
package inheritance;
public class Literature {
  protected String title;
  protected String author;

  public Literature(String title, String author)
  {
     this.title = title;
     this.author = author;
   }
```

4.  当我们开始时，让我们给`Literature`一个类似的`Print`方法，就像我们分配给`Book`和`Poem`类的方法一样：

```java
public void Print()
{
   System.out.println(title);
   System.out.println("\tWritten By: " + author);
 }
```

现在，如果我们愿意，我们可以转到`main`方法并声明`Literature`类的对象，但这不是重点。这不是我们创建`Literature`类的原因。相反，我们的目标是利用这个`Literature`类作为基础，我们将在此基础上声明更具体的文学类型，如诗歌或书籍。为了利用我们的`Literature`类，让我们看看它如何应用于现有的`Poem`类。

# 这是一种关系

我们的`Literature`类包含管理一篇文献的标题和作者所需的声明和所有功能。如果我们让 Java 知道`Poem`和`Literature`之间存在继承关系，我们应该能够从以下`Poem`类的标题和作者中删除所有真实引用：

```java
package inheritance;
public class Poem extends Literature{
    private String title;
    private String author;
    private String style;
```

首先，让我们谈谈我们修改过的`Poem`类的声明。当我们说一个类扩展另一个类时，我们是说它们之间存在一种 is-a 关系，这样我就可以逻辑地说，“一首诗就是一篇文学作品。”用更多的 Java 术语来说，`Poem`子类扩展或继承了`Literature`类。这意味着当我们创建一个`Poem`对象时，它将拥有它扩展的类的所有成员和功能：

```java
package inheritance;
public class Poem extends Literature {
    private String style;

    public Poem(String title, String author, String style)
```

在我们的例子中，其中两个成员是`title`和`author`。`Literature`类声明这些成员，并在整个类的功能中很好地管理它们。因此，我们可以从`Poem`类中删除这些成员，我们仍然可以在`Poem`类的方法中访问它们。这是因为`Poem`类只是从`Literature`继承了它的声明。不过，我们需要做一些轻微的修改。当我们构造从另一个类继承的类的对象时，默认情况下，子类的构造函数（如前面的屏幕截图所示）将首先调用超类的构造函数：

```java
package inheritance;
public class Literature {
    protected String title;
    protected String author;

    public Literature(String title, String author)
    {
         this.title = title;
         this.author = author;
    }
```

这让 Java 感到困惑，因为我们现在已经设置好了它，因为`Poem`构造函数需要三个变量作为输入，而`Literature`构造函数只需要两个变量。要解决此问题，请使用以下步骤显式调用`Poem`构造函数中的`Literature`构造函数：

1.  当我们在一个子类中时，我们可以使用`super`关键字调用我们的超类的方法。在本例中，我们将通过调用`super`构造函数或`Literature`构造函数开始`Poem`构造函数，并向其传递我们希望它了解的属性：

```java
public Poem(String title, String author, String style)
{
     super(title, author);
     this.style = style;
 }
```

2.  我们可以在`Print`方法中做一些非常类似的事情，因为我们的`Literature`类，我们的超类，已经知道如何打印标题和作者。`Poem`类没有理由实现此功能：

```java
 public void Print()
 {
      super.Print();
      System.out.println("\tIn The Style Of: " + style);
 }
```

如果我们通过调用`super.Print`开始`Print`方法，而不是前面屏幕截图中显示的原始显式打印行，那么我们将从`Print`方法获得相同的行为。现在，当 Poem 的`Print`方法运行时，它将首先调用超类的，即`Literature.java`类的`Print`方法。它将最终打印出`Poem`课程的风格，这不是所有文学作品都能共享的。

虽然我们的`Poem`构造函数和`Literature`构造函数有不同的名称，甚至有不同的输入样式，但`Poem`和`Literature`之间共享的两个`Print`方法完全相同。稍后我们将进一步讨论这个问题，但现在您应该知道，我们在这里使用了一种称为**覆盖**的技术。

# 最重要的

当我们声明一个子类有一个与它的一个超类方法相同的方法时，我们已经重写了超类方法。当我们这样做时，最好使用 Java`Override`指示器：

```java
@Override public void Print()
```

这是对未来的程序员和我们的编译套件中一些更神秘的元素的一个指示，在前面的屏幕截图中给出了这个特定的方法，它隐藏了一个方法。当我们实际运行代码时，Java 会优先选择方法的最低版本或子类版本。

那么，让我们看看我们是否成功地宣布了我们的`Poem`和`Literature`关系。让我们回到`Inheritence.java`类中程序的`main`方法，看看这个程序的诗部分是否像以前一样执行：

![](img/a8018c55-7173-4c5e-aa28-0bf9b697cc82.png)

当我们运行这个程序时，我们得到了与之前完全相同的输出，这是一个很好的迹象，表明我们已经将`Poem`类设置为以逻辑方式从`Literature`继承。

现在我们可以跳到我们的`Book`课了。我们将使用以下步骤将其设置为`Book`和`Literature`类之间的 is-a 关系：

1.  首先，我们将声明`Book`扩展了`Literature`类；然后，我们将在`Book`类中删除对 title 和 author 的引用，因为`Literature`类，即超类，将处理以下内容：

```java
        package inheritance;
        public class Book extends Literature{
        private String publisher;
        private String genre;
```

2.  与`Poem`类一样，我们需要显式调用`Literature`类的构造函数并将`title`和`author`传递给它：

```java
        public Book(String title, String author, String publisher, String
        genre)
        {
             super(title, author);
             this.publisher = publisher;
             this.genre = genre;
         }
```

3.  然后，我们可以利用超类的`Print`方法简化`Book`类的打印：

```java
        @Override public void Print()
        {
             super.Print();
             System.out.println("\tPublished By: " + publisher);
             System.out.println("\tIs A: " + genre);
```

4.  Once again, let's jump back to our `main` method and run it to make sure that we've done this successfully:

    ![](img/269c7dd3-0d67-4a9f-99cd-9bdae0783358.png)

我们开始：`The Lord of the Rings`输出，就像我们以前看到的一样。就风格而言，这种变化真的很棒。通过添加`Literature`类，然后将其子类化以创建`Book`和`Poem`类，我们已经使`Book`和`Poem`类变得更容易理解，程序员也更容易理解发生了什么。

然而，这种变化并不是纯粹的风格上的。通过声明`Book`、`Poem`和`Literature`类之间的 is-a 关系，其中`Book`和`Poem`继承自`Literature`类，我们为自己提供了以前没有的实际功能。让我们来看看这个动作。如果我们回到我们的`main`方法，让我们假设，我们处理的不是单个`Book`和`Poem`类，而是需要存储在某种数据结构中的庞大网络。对于我们最初的实现，这将是一个真正的挑战。

# 数据结构

没有一种易于访问的数据结构可以同时存储书籍和诗歌。我们可能必须使用两种数据结构或打破强类型，这是 Java 的全部要点：

```java
Book[] books = new Book[5];
```

然而，在我们的新实现中，`Book`和`Poem`都继承自`Literature`，我们可以将它们存储在相同的数据结构中。这是因为继承是一种“是”与“是”的关系，这意味着一旦我们继承了某样东西，我们就可以提出诸如书是文学，诗也是文学之类的主张。如果这是真的，那么`Literature`对象数组应该能够在其中存储`Book`和`Poem`。让我们通过以下步骤来说明这一点：

1.  创建一个`Literature`对象数组：

```java
 Literature[] lits = new Literature[5];
 lits[0] = a;
 lits[1] = b;
```

我们在构建这个项目时没有遇到编译错误这一事实是一个很好的迹象，表明我们正在做一些合法的事情。

2.  为了便于演示，让我们在这里充实我们的数组，以包含书籍和诗歌的数量：

```java
 Literature[] lits = new Literature[5];
 lits[0] = a;
 lits[1] = b;
 lits[2] = a;
 lits[3] = b;
 lits[4] = a;
```

我们将修改我们的`main`方法，直接从数组中打印出来。现在，当我们使用我们的子类时，就好像它们是它们的超类的对象一样，我们必须意识到我们现在引用它们作为那个超类的对象。例如，当我们通过`Literature`数组获取一个元素时，无论该元素是否为`Book`类，我们仍然无法访问其`genre`字段等内容，即使该字段为`public`：

```java
 Literature[] lits = new Literature[5];
 lits[0] = a;
 lits[1] = b;
 lits[2] = a;
 lits[3] = b;
 lits[4] = a;
 for(int i=0; i< lits.length; i++)
 {
      lits[i].Print(); 
 }
```

这是因为我们现在用作对象的`Literature`类（如前面的屏幕截图所示）没有`genre`成员变量。但是我们可以做的是从我们的超类调用被子类覆盖的方法。

3.  我们可以在`for`循环中调用`Literature`类的`Print`方法。Java 将优先考虑我们子类的`Print`方法：

```java
for(int i=0; i< lits.length; i++)
{
     lits[i].Print(); 
 }
```

这意味着，当我们运行这个程序时，我们仍然会得到我们归因于`Book`和`Poem`的特殊格式的输出，而不是我们存储在`Literature`类中的简化版本：

```java
public void Print()
{
     System.out.println(title);
     System.out.println("\tWritten By: " + author);
 }
```

# 抽象方法

我们有时可能会看到存在的方法只会被子类重载。这些方法什么都不做，我们可以在超类（`Literature.java`中用`abstract`关键字标记它们，即`public abstract void Print()`。当然，如果一个类声明了方法`abstract`，这可能是一个很好的迹象，表明永远不应该显式创建这样一个类的实例。如果我们的`Literature`类的`Print`方法是抽象的，那么我们永远不应该声明只属于文献的对象。我们应该只使用属于`Literature`子类之一的对象。如果我们要沿着这条路线走，我们也应该宣布`Literature`为`abstract`类：

```java
package inheritance;
public abstract class Literature {
```

如果我们这样做了，当然，我们必须去掉对文学类超级方法的引用，所以现在让我们撤销这些更改。

让我们先来看看这个程序最初所犯的一个小错误。在创建文学类时，我们声明标题和作者是`public`成员变量。您可能知道，如果没有充分的理由，通常我们不会将成员变量声明为公共。对于一篇文献来说，在它已经被声明之后改变它的作者是没有多大意义的，所以`author`和`title`应该是`Literature`类的构造函数中设置的`private`成员变量，其值永远不应该改变。不幸的是，如果我们对文学课做了这样的改变，我们将限制诗歌和书籍课的功能。

比如说，我们想修改`Poem`类的`Print`函数，这样它就不必显式调用`Literature`类的`Print`函数：

```java
@Override public void Print()
{
     System.out.println(title);
     System.out.println("\tWritten By: " + author);
     System.out.println("\tIn The Style Of: " + style);
 }
```

也许我们想先告诉它我们在这里声明一个`Poem`类：

```java
System.out.println("POEM: " + title);
```

不幸的是，由于我们将`title`和`author`设置为`Literature`类的私有，`Poem`类即使是`Literature`的子类，也无法在显式代码中访问这些成员变量。这有点烦人，似乎在`private`和`public`之间有某种保护设置，除了类的子类之外，每个类都是私有的。事实上，有一个可用的保护设置。

# 保护方法

`protected`方法为保护设置。如果我们声明成员变量为`protected`，则意味着它们是私有的，除类及其子类外，任何人都无法访问：

```java
package inheritance;
public class Literature {
    protected String title;
    protected String author;
```

让我们自己放心，我们在这里所做的一切都是合法的。让我们再次运行我们的程序，并确保输出看起来良好，情况就是这样。在这之后，我们应该对继承有一个相当好的把握。我们可以开发许多真正模仿真实世界的系统，我们可以使用继承和小型类编写非常优雅的功能性代码，这些类本身不会做很多复杂的事情。

# 摘要

在本节中，我们将快速了解一个与 Java 中的继承相关的重要思想。为了让我们了解我们将要讨论的内容，我认为最好是在系统中启动一个现有的项目。让我们来看一下代码文件中的代码。

到目前为止，我们已经完成了以下工作：

*   我们程序的`main`方法创建一个对象列表。这些对象属于`Book`或`Poem`类型，但我们将它们放在`Literature`对象列表中，这让我们相信`Book`和`Poem`类必须继承或扩展`Literature`类。
*   一旦我们构建了这个数组，我们只需使用`for`循环遍历它，并在每个对象上调用`for`循环的`Print`方法。
*   在这一点上，我们将对象视为`Literature`对象，而不是它们处于最低层次的书籍或诗歌。这让我们相信`Literature`类本身必须实现`Print`方法；如果我们跳进课堂，我们会发现这是真的。

然而，如果我们运行我们的程序，我们将很快看到书籍和诗歌执行它们的`Print`方法略有不同，为它们各自显示不同的信息。当我们看`Book`和`Poem`类时，我们可以解释这一点，它们确实扩展了`Literature`类，但是这些类中的每一个都覆盖`Literature`类的`Print`方法来提供自己的功能。这一切都很好，是一个非常优雅的解决方案，但有一个有趣的案例，我们应该看看并讨论。因为`Literature`本身就是一个类，所以我们可以声明一个新的`Literature`对象，就像我们可以声明`Book`或`Poem`一样。`Literature`类的构造函数首先期望文献的`title`，然后是`author`。一旦我们创建了`Literature`类的新实例，我们就可以将该实例放入`Literature`类列表中，就像我们对`Book`和`Poem`类的实例所做的那样：

```java
Literature l= new Literature("Java", "Zach");
Literature[] lits = new Literature[5];
lits[0] = a;
lits[1] = b;
lits[2] = l;
lits[3] = b;
lits[4] = a;
for(int i=0; i< lits.length; i++)
{
     lits[i].Print(); 
 }
```

当我们这样做并运行程序时，我们将看到`Literature`类的`Print`方法被执行，我们创建的新`Literature`对象将显示在我们的书籍和诗歌列表旁边：

![](img/a451f7d1-909e-45b3-a3bc-52665ed6e47b.png)

那么这里有什么问题？好吧，这取决于我们试图设计的软件的真实性质，这可能有意义，也可能没有意义。假设我们是作为图书馆系统的一部分来做这件事的，只向某人提供一个叫做 Java 的东西是由一个叫 Zach 的人写的，而不告诉他们这是一本书还是一首诗，或者我们决定与特定类型的文学相联系的任何其他信息。这可能根本没有用，而且是永远不应该做的事情。

如果是这样的话，Java 为我们提供了一个创建类的系统，我们可以将这些类用于继承目的，但我们永远无法像以前那样合法地自己实例化它们。如果我们想将一个类标记为该类型，我们将其称为`abstract`类，在该类的声明中，我们只需使用`abstract`关键字：

```java
public abstract class Literature {
```

一旦我们标记了一个类`abstract`，实例化这个类就不再是合法的操作。从表面上看，这是一件非常简单的事情，主要是“保护我们的代码不受我们自己和其他程序员的影响”，但这并不完全正确；相反，这是正确的，但它不是将类声明为`abstract`的唯一目的。

一旦我们告诉 Java，我们永远不能只创建`Literature`的实例，即使用`Literature`作为超类的类的实例，我们在设置`Literature`类时就不再受到限制。因为我们已经声明了`Literature`是一个抽象类，所以我们和 Java 都知道`Literature`永远不会单独被实例化，只有当它是被实例化的类的超类时。在这种情况下，我们可以不使用这个类中大多数 Java 类必须拥有的部分。例如，我们实际上不需要为`Literature`声明构造函数。如果`Literature`是一个标准的 Java 类，Java 就不会同意这样做，因为如果我们尝试实例化`Literature`，它将不知道该做什么。没有可调用的构造函数。但由于`Literature`是抽象的，我们可以确信文学的子类将有自己的构造函数。当然，如果我们做了这样的更改，我们必须在我们的子类中去掉对`Literature`构造函数的引用，也就是说，从子类中删除`super`方法。因此，这一变化肯定有一个权衡。它需要我们的子类中有更多的代码，才能在`Literature`超类中有更少的代码。在这个特殊的例子中，这种折衷可能不值得，因为我们在复制`Book`和`Poem`构造函数之间的代码，但是如果可以假设`Literature`子类的构造函数做了非常不同的事情，那么不声明公共基构造函数就很有意义了。

因此，简言之，当我们构建程序或更大的解决方案时，我们应该将这些类声明为`abstract`，这样我们的类对于架构目的来说非常有意义，但永远不应该单独创建。有时候，当某些常见的类功能（比如有构造函数）对这个类没有意义时，我们就会知道我们遇到了这样的类。

# 总结

在本章中，我们通过创建一个称为超类和子类的东西并在它们之间建立“is-a”关系，精确地使用继承的概念，了解了面向对象编程的一些复杂之处。我们还讨论了一些关键方面的用法，例如重写子类和超类、数据结构和`protected`方法。我们还详细了解了`abstract`方法的工作原理。

在下一章中，您将学习有用的 Java 类。